""" Extension fields for MLSchema, specifically for datetime \
which doesn't work right now from yaml """
from datetime import datetime

from marshmallow import fields

from uuid import UUID


# pylint: disable=missing-class-docstring, too-few-public-methods
class MLSchemaFields:

    ALL_FIELD_TYPES = {
        "Raw": "raw",
        "Nested": "nested",
        "Dict": dict,
        "List": list,
        "Tuple": tuple,
        "String": str,
        "UUID": UUID,
        "Integer": int,
        "Boolean": bool,
        "Float": float,
        "DateTime": datetime,
        "Url": "url",
        "URL": "uri",
        "Email": "email",
        "Str": str,
        "Bool": bool,
        "Int": int,
    }

    class MLField(fields.Field):
        """ Subclassing of Marshmallow field for everything we want to extend in MLSchema """

        identity_type: type

        def _deserialize(self, value, attr, data, **kwargs):
            if isinstance(value, self.identity_type):
                return value
            return super()._deserialize(value, attr, data, **kwargs)

    class DateTime(MLField, fields.DateTime):
        identity_type = datetime



from mlspeclib.mlobject import MLObject
from mlspeclib.mlschema import MLSchema
from mlspeclib.mlschemaenums import MLSchemaTypes
from mlspeclib.mlschemavalidators import MLSchemaValidators
from mlspeclib.experimental.metastore import Metastore
from mlspeclib.experimental.gremlin_helpers import GremlinHelpers

# from mlspeclib.helpers import MLObjectHelpers



""" Helper files for all functions """

from io import StringIO
import yaml as YAML
import json as JSON
import uuid
import ast

from mlspeclib.mlschemaenums import MLSchemaTypes
from mlspeclib.mlschemafields import MLSchemaFields

from marshmallow.fields import Field, ValidationError
from marshmallow.class_registry import RegistryError

import base64
import pickle

ALLOWED_OPERATORS = ["<", "<=", ">", ">=", "==", "%", "<>", "!="]


def repr_uuid(dumper, uuid_obj):
    return YAML.ScalarNode("tag:yaml.org,2002:str", str(uuid_obj))


def convert_yaml_to_dict(value):
    """ Converts raw text to yaml using ruamel (put into a helper to ease
        converting to other libraries in the future) """

    if isinstance(value, dict):
        return value
    else:
        return YAML.safe_load(value)


def convert_dict_to_yaml(value):
    """ Converts dict to yaml using ruamel (put into a helper to ease
        converting to other libraries in the future) """
    if isinstance(value, str):
        return value

    # pylint: disable=line-too-long
    string_io_handle = StringIO()
    YAML.SafeDumper.add_representer(uuid.UUID, repr_uuid)
    YAML.safe_dump(value, string_io_handle)
    return string_io_handle.getvalue()


def merge_two_dicts(first_dict, second_dict):
    """ Merges two python dicts by making a copy of first then updating with second.
        Returns a copy. """

    return_dict = first_dict.copy()  # start with x's keys and values
    return_dict.update(
        second_dict
    )  # modifies z with y's keys and values & returns None
    return return_dict


def check_and_return_schema_type_by_string(val):
    """ Looks up string in mlspeclib.mlschemaenums and returns enum of type SchemaTypes """

    return val

    # if isinstance(val, MLSchemaTypes):
    #     return val

    # try:
    #     return MLSchemaTypes[val.upper()]
    # except AttributeError:
    #     raise KeyError("'%s' is not an enum from MLSchemaTypes" % val)
    # except KeyError:
    #     raise KeyError("'%s' is not an enum from MLSchemaTypes" % val)


def recursive_fromkeys(full_dict: dict):
    """ Builds a new dict with no values in it. Works recursively, but only looks for objects \
        with the 'nested' attribute."""
    return_dict = {}
    for key in full_dict.keys():
        if hasattr(full_dict[key], "nested"):
            return_dict[key] = recursive_fromkeys(
                full_dict[key].nested._declared_fields
            )
        else:
            if isinstance(full_dict[key], dict):
                return_dict[key] = {}
            else:
                return_dict[key] = None

    return return_dict


def contains_minimum_fields_for_schema(schema_dict: dict) -> bool:
    return_val = True
    return_val &= "mlspec_schema_version" in schema_dict
    # return_val &= "mlspec_base_type" in schema_dict
    return_val &= "mlspec_schema_type" in schema_dict
    return_val &= "schema_version" in schema_dict
    return_val &= "schema_type" in schema_dict

    # Check for the sub values after we've already done the basics
    if return_val:
        return_val &= "meta" in schema_dict["mlspec_schema_version"]
        # return_val &= "meta" in schema_dict["mlspec_base_type"]
        return_val &= "meta" in schema_dict["mlspec_schema_type"]
    return return_val


def valid_comparison_operator(val):
    return val in ALLOWED_OPERATORS


def generate_lambda(user_submitted_string):
    # make a list of safe functions
    safe_list = ["math", "lambda"]

    # use the list to filter the local namespace
    safe_dict = dict([(k, locals().get(k, None)) for k in safe_list])

    all_args = set()

    try:
        node = ast.parse(user_submitted_string, mode="eval")
    except TypeError:
        raise ValidationError(f"No value was passed in as a function as a constraint.")
    except SyntaxError:
        raise ValidationError(
            f'No parsable lambda was detected. Try it yourself: `ast.parse({user_submitted_string}, mode="eval")`'
        )

    for elem in ast.walk(node):
        if isinstance(elem, ast.Name):
            all_args.update(str(elem.id))

    if len(all_args) > 1:
        raise ValidationError(
            f"Only one variable ('x') is supported in lambda constraints at this time. The following variables were detected: '{','.join(all_args)}'"
        )

    if len(all_args) == 0:
        raise ValidationError(
            f"No variables were detected in the lambda constraint: '{user_submitted_string}'"
        )

    if list(all_args)[0] != "x":
        raise ValidationError(
            f"Only the variable 'x' is supported at this time. The following variables were detected: '{','.join(all_args)}'"
        )

    lambda_string = f"lambda {','.join(all_args)}: {user_submitted_string}"
    return_lambda = eval(lambda_string, {"__builtins__": None}, safe_dict)
    if lambda_string is None:
        raise ValidationError(
            'Could not parse %s into a lambda with one variable. Test yourself by running this on the command line: \
            \'eval(%s, {"__builtins__": None}, %s'
            % user_submitted_string,
            lambda_string,
            str(safe_dict),
        )
    else:
        return return_lambda


# Functions below here are just helper functions for building names.
def build_schema_name_for_schema(
    mlspec_schema_version: str, mlspec_schema_type: str, schema_prefix: str = None
):
    """ Generates schema name based on the fields in the dict.
    Moved to a helper function to ensure consistency. """

    try:
        mlspec_schema_type_string = mlspec_schema_type["meta"]
    except KeyError:
        raise KeyError("No mlschema_schema_type provided.")

    try:
        mlspec_schema_version_string = mlspec_schema_version["meta"]
    except KeyError:
        raise KeyError("No mlschema_schema_version provided.")

    schema_name = return_schema_name(
        mlspec_schema_version_string, mlspec_schema_type_string, schema_prefix
    )

    return schema_name


def build_schema_name_for_object(
    schema_object: dict = None, submission_data: dict = None, schema_prefix: str = None,
):
    """ Retrieves a schema_name from either the schema_object or the submitted data. """

    if schema_object is None and submission_data is None:
        raise KeyError(f"Neither schema_object nor submission_data was provided.")

    if (schema_object is not None and hasattr(schema_object, "schema_name")) and (
        schema_object.schema_name is not None
    ):
        schema_name = schema_object.schema_name
    elif "schema_name" in submission_data:
        schema_name = submission_data["schema_name"]
    elif (
        "schema_type" in submission_data and "schema_version" in submission_data
    ) and (
        submission_data["schema_type"] is not None
        and submission_data["schema_version"] is not None
    ):
        schema_name = return_schema_name(
            submission_data["schema_version"],
            submission_data["schema_type"],
            schema_prefix,
        )
    else:
        raise KeyError(
            f"Not enough information submitted to build a schema \
                            name for submission to class_registry."
        )

    return schema_name


def return_schema_name(
    raw_schema_version_string: str,
    raw_schema_type_string: str,
    schema_prefix: str = None,
):
    """ Takes Schema Version and Schema Type and returns a transformed schema name.
    Optionally takes a schema prefix to attach to the front. """

    schema_version_string = raw_schema_version_string.replace("-", r"_").replace(
        ".", r"_"
    )
    schema_name = schema_version_string + "_" + raw_schema_type_string.lower()

    if schema_prefix and schema_name:
        schema_name = schema_prefix + "_" + schema_name

    return schema_name


def get_sub_schema_name(schema_name, field_name):
    return schema_name + "_" + field_name.lower()


def encode_raw_object_for_db(mlobject):
    # Converts object -> dict -> yaml -> base64
    dict_conversion = mlobject.dict_without_internal_variables()
    yaml_conversion = convert_dict_to_yaml(dict_conversion)
    encode_to_utf8_bytes = yaml_conversion.encode("utf-8")
    base64_encode = base64.urlsafe_b64encode(encode_to_utf8_bytes)
    final_encode_to_utf8 = str(base64_encode, "utf-8")
    return final_encode_to_utf8


def decode_raw_object_from_db(s: str):
    # Converts base64 -> yaml
    base64_decode = base64.urlsafe_b64decode(s)
    return base64_decode


def to_yaml(this_dict: dict):
    return YAML.dump(this_dict)


def to_json(this_dict: dict):
    return JSON.dump(this_dict)


# def convert_marshmallow_field_to_primitive(marshmallow_field: Field):
#     field_name = type(marshmallow_field).__name__
#     try:
#         return MLSchemaFields.ALL_FIELD_TYPES[field_name]
#     except KeyError:
#         raise KeyError(f"'{field_name}' is not a known field type for code generation.")



""" All allowed schema types as enums."""
from enum import Enum, auto

# TODO: Name spacing across schema - path to the definition library
# (folder than just has a list of definitions)
# User experience is at the top of a notebook say "mlspeclib.add_schema_registry(file_path)"
# TODO: Hierarchy of name lookups - two schema registries, need to start with root and cascade down
# - can reorder in lookup in schema registry


# TODO: Build this dynamically (maybe?) from files in mlspeclib/data
class MLSchemaTypes(Enum):
    """ Types of allowed schemas, the numbers don't matter so we populated them
    using auto()."""
    BASE = auto()
    DATAPATH = auto()
    ENVIRONMENT = auto()
    CONTAINERSTORE = auto()
    CONVERSION = auto()
    DATASCHEMA = auto()
    DATASOURCE = auto()
    METADATA = auto()
    SCORE = auto()
    SERVE = auto()
    TRAIN_EXECUTION = auto()
    TRAIN_RESULTS = auto()
    TRAINING_POST_PROCESS = auto()
    RUNCONFIG = auto()
    MODEL = auto()
    JOB = auto()
    COMPONENT = auto()



# pylint: disable=attribute-defined-outside-init
""" Functions for loading and saving files to disk. """
import semver
from box import Box
import datetime
from pathlib import Path

import marshmallow.class_registry
from marshmallow import ValidationError

from mlspeclib.io import IO
from mlspeclib.mlschema import MLSchema
from mlspeclib.mlschemaenums import MLSchemaTypes
from mlspeclib.helpers import (
    check_and_return_schema_type_by_string,
    recursive_fromkeys,
    convert_yaml_to_dict,
    return_schema_name,
    to_yaml as helpers_to_yaml,
    to_json as helpers_to_json,
)


class Printable_Field:
    printable_name = None
    required = False
    field_type = None
    description = None
    constraint = None

    def __init__(
        self,
        printable_name: str,
        field_type,
        required: bool = False,
        description: str = None,
        constraint: str = None,
    ):
        self.printable_name = printable_name
        self.required = required
        self.field_type = field_type
        self.description = description
        self.constraint = constraint


class MLObject(Box):
    """ Contains all the fields loaded from an MLSpec, and validated against the MLSchema. Also
    provides load and save functions."""

    __version='0.0.1'

    def set_type(self, schema_version, schema_type, schema=None, schema_object=None):
        """ Used primarily after MLObject instantiation to set the schema_version and
        schema_type. Does verification of both fields and then loads a stub object
        with all fields set to 'None' except schema_version and schema_type."""

        try:
            semver.VersionInfo.parse(schema_version)
            self.__schema_version = schema_version
        except ValueError:
            raise ValueError(f"'{schema_version}' is not a valid semantic version.")
        except TypeError:
            raise ValueError(f"'{schema_version}' is not a valid semantic version.")

        self.__schema_type = check_and_return_schema_type_by_string(schema_type)
        self.__schema = schema
        self.__schema_object = schema_object

        self.create_stub_object()
        self.schema_version = self.get_schema_version()
        self.schema_type = self.get_schema_type().name.lower()

    def set_semver_and_type(
        self,
        schema_version=None,
        schema_type: MLSchemaTypes = None,
        contents_as_dict=None,
    ):
        """ Mostly internal function used to set variables for version and type. Should only \
            use rarely - if you use this separate from loading an object, you can get out \
            of sync."""

        if (schema_version is not None and schema_type is not None) and (
            contents_as_dict is not None
        ):
            raise Warning(
                "Semantic version + schema type and a dictionary were \
                           provided as the version & schema type, defaulting to \
                           using the content from the dictionary."
            )

        if contents_as_dict is not None:
            try:
                schema_version = contents_as_dict["schema_version"]
            except KeyError:
                raise KeyError(
                    "No field named 'schema_version' found at the top level of data."
                )

            try:
                schema_type = contents_as_dict["schema_type"]
            except KeyError:
                raise KeyError(
                    "No field named 'schema_type' found at the top level of data."
                )

    def create_stub_object(self):
        """ Creates a stub dictionary based on the schema with all values set to None.
        We do this because our goal is to prevent (eventually) creation of new attributes
        directly by users - only allowing them to use what we already provide to them
        to keep the schema in sync with the object."""

        MLSchema.populate_registry()
        version_number = self.get_schema_version()
        self.__schema_name = return_schema_name(
            version_number, self.get_schema_type().name
        )
        object_schema = marshmallow.class_registry.get_class(self.get_schema_name())
        self.__schema = object_schema()
        self.__schema_object = None
        these_fields = object_schema().fields
        this_key_dict = recursive_fromkeys(these_fields)
        self.merge_update(this_key_dict)

    def validate(self):
        """ Returns an array of errors after validating against the assigned
        schema. Each error is an array with the message at the 0-th index."""
        return self.get_schema().validate(self.dict_without_internal_variables())

    def save(self, save_path: Path, export_json=False):
        """ Copies all internal data to an MLSchema, which it then
        uses to validate, and saves to disk. Returns True on success,
        or False and an array of errors if it fails schema validation. """
        errors = self.validate()

        file_write_success = False
        if len(errors) == 0:
            file_path = save_path / (
                self.get_schema_name()
                + "-"
                + datetime.datetime.now().isoformat()
                + ".yaml"
            )

            export_dict = self.dict_without_internal_variables()
            if not export_json:
                IO.write_content_to_path(file_path, export_dict)
            else:
                IO.write_content_to_path(file_path, export_dict)

            # Expecting the file system to throw an error if something went wrong above.
            # By this point, the file system has written and so recording the filename.
            self.__file_path = file_path
            file_write_success = True
        return (file_write_success, errors)

    def to_yaml(self):
        return helpers_to_yaml(self.dict_without_internal_variables())

    def to_json(self):
        return helpers_to_json(self.dict_without_internal_variables())

    @staticmethod
    def code_gen(
        schema_version, schema_type, type_hints=True, prefix=None, only_required=False
    ) -> None:
        print(
            MLObject._code_gen_string(
                schema_version,
                schema_type,
                prefix=prefix,
                only_required=False,
                type_hints=type_hints,
            )
        )

    @staticmethod
    def _code_gen_string(
        schema_version, schema_type, prefix=None, only_required=False, type_hints=True
    ) -> str:
        stub_object = MLObject()
        stub_object.set_type(schema_version=schema_version, schema_type=schema_type)
        if prefix is None:
            prefix = "VARIABLE_NAME"
        all_objects = MLObject._build_all_printable_objects(
            stub_object.get_schema()._declared_fields, prefix
        )

        required_string = ""
        optional_string = ""

        for printable_object in all_objects:
            if printable_object.required:
                required_string += MLObject._string_for_field(
                    printable_object, type_hints
                )
            else:
                optional_string += MLObject._string_for_field(
                    printable_object, type_hints
                )

        complete_string = f"""

{prefix} = MLObject()
{prefix}.set_type('{schema_version}', '{stub_object.get_schema_type().name.lower()}')

#
# All required attributes
#
{required_string}

#
# All optional attributes
#
{optional_string}"""

        return complete_string

    @staticmethod
    def _string_for_field(printable_object: Printable_Field, type_hints=True) -> str:
        description_string = ""
        if printable_object.description is not None:
            description_string = f"""# {printable_object.description}
"""
        type_hint_string = ""
        if type_hints:
            type_hint_string = f"""\n# {printable_object.printable_name} expects -> {printable_object.field_type}
"""
            if printable_object.constraint is not None:
                type_hint_string += f"\n# Constraint: {printable_object.constraint}"

        assignment_string = f"""{printable_object.printable_name} =
"""
        return f"{description_string}{type_hint_string}{assignment_string}" ""

    @staticmethod
    def _build_all_printable_objects(fields_dict: dict, prefix: str) -> []:
        return_array = []
        for key in fields_dict:
            if key == "schema_version" or key == "schema_type":
                continue

            field = fields_dict[key]

            if hasattr(field, "nested"):
                return_array.extend(
                    MLObject._build_all_printable_objects(
                        field.schema._declared_fields, f"{prefix}.{key}"
                    )
                )
            else:
                description = None
                if hasattr(field, "description"):
                    description = field.description

                constraint = None
                if hasattr(field, "constraint"):
                    constraint = field.constraint

                return_array.append(
                    Printable_Field(
                        printable_name=f"{prefix}.{key}",
                        required=field.required,
                        field_type=type(field).__name__,
                        description=description,
                        constraint=constraint,
                    )
                )
        return return_array

    @staticmethod
    def create_object_from_file(file_path: str):
        """ Creates an MLObject based on a file path. File must be valid yaml."""
        ml_content_from_disk = IO.get_content_from_path(file_path)
        return MLObject.create_object_from_string(ml_content_from_disk)

    @staticmethod
    def create_object_from_string(file_contents):
        """ Creates an MLObject based on a string. String must be valid yaml.
        Returns Tuple MLObject and list of errors."""
        if(isinstance(file_contents, dict)):
            contents_as_dict = file_contents
        else:
            contents_as_dict = convert_yaml_to_dict(file_contents)

        # if (self.schema_type() is not None and self.version() is not None):
        #     schema_string = self.schema_type().name.lower()
        #     if (self.version() != contents_as_dict['schema_version'] or \
        #         schema_string != contents_as_dict['schema_type']):
        #         raise AttributeError("""The schema version and schema type were not in sync
        #     with those provided in the data. Rather than guessing which you want to use, we
        #     are throwing this error:
        #     Version Expected: %s
        #     Version Provided: %s
        #     Schema Type Expected: %s
        #     Schema Type Provided: %s """ % (self.version(), contents_as_dict['schema_version'], \
        #                                 schema_string, contents_as_dict['schema_type']))

        ml_object = MLObject()
        ml_object.set_type(
            schema_version=contents_as_dict["schema_version"],
            schema_type=contents_as_dict["schema_type"],
        )
        MLObject.update_tree(ml_object, contents_as_dict)
        errors = ml_object.validate()

        if len(errors) > 0:
            return (None, errors)
        else:
            return (ml_object, {})

    @staticmethod
    def update_tree(object_to_update, content):
        """ Updates the box tree to the content provided, and does a validation at the end. """
        for key in object_to_update:
            if isinstance(object_to_update[key], Box) and key in content.keys():
                MLObject.update_tree(object_to_update[key], content[key])

        object_to_update.merge_update(content)
        return object_to_update

    def dict_without_internal_variables(self):
        """ Returns a dict of all values on MLObject minus any with the prefix '_MLObject'
        (which are only for internal use). This allows us to validate on the whole dict."""
        return {
            k: v for k, v in self.to_dict().items() if not k.startswith("_MLObject")
        }

    # Simple getter functions with a prefix so that it's unlikely to be hidden
    # by values in the schema.

    # pylint: disable=missing-function-docstring
    def get_schema_version(self):
        return self.__schema_version

    # pylint: disable=missing-function-docstring
    def get_schema_type(self):
        if isinstance(self.__schema_type, MLSchemaTypes):
            return self.__schema_type
        else:
            # TODO: The below grossness is because I don't feel like going around and removing all the
            # places where get_schema_type().name is used. If I did, I could remove almost everything
            # here.
            class Object(object):
                pass

            a = Object()
            a.name = self.__schema_type
            return a

    # pylint: disable=missing-function-docstring
    def get_schema(self):
        return self.__schema

    # pylint: disable=missing-function-docstring
    def get_schema_object(self):
        return self.__schema_object

    # pylint: disable=missing-function-docstring
    def get_schema_name(self):
        return self.__schema_name

    # pylint: disable=missing-function-docstring
    def get_file_path(self):
        return self.__file_path
